@@ -111,7 +111,7 @@ extern long cm_ForceUpdateVolume(struct cm_fid *fidp, cm_user_t *userp,
 	cm_req_t *reqp);
 
 extern cm_serverRef_t **cm_GetVolServers(cm_volume_t *volp, afs_uint32 volume,
-                                         cm_user_t *userp, cm_req_t *reqp);
+                                         cm_user_t *userp, cm_req_t *reqp, afs_uint32 *replicated);
 
 extern void cm_ChangeRankVolume(cm_server_t *tsp);
 
@@ -163,5 +163,10 @@ extern cm_vol_state_t * cm_VolumeStateByName(cm_volume_t *volp, char *volname);
 
 extern afs_int32 cm_VolumeType(cm_volume_t *volp, afs_uint32 id);
 
+extern long cm_UpdateVolumeLocation(struct cm_cell *cellp, cm_user_t *userp, cm_req_t *reqp,
+                                    cm_volume_t *volp);
+
+extern afs_int32 cm_IsVolumeReplicated(cm_fid_t *fidp);
+
 extern osi_rwlock_t cm_volumeLock;
 #endif /*  OPENAFS_WINNT_AFSD_CM_VOLUME_H */
