@@ -109,7 +109,7 @@ AFSRetrieveAuthGroup( IN ULONGLONG ProcessId,
 
         AFSDbgLogMsg( AFS_SUBSYSTEM_LOCK_PROCESSING,
                       AFS_TRACE_LEVEL_VERBOSE,
-                      "AFSRetrieveAuthGroup Acquiring Control ProcessTree.TreeLock lock %08lX SHARED %08lX\n",
+                      "AFSRetrieveAuthGroup Acquiring Control ProcessTree.TreeLock lock %p SHARED %08lX\n",
                       pDeviceExt->Specific.Control.ProcessTree.TreeLock,
                       PsGetCurrentThread());
 
@@ -135,6 +135,7 @@ AFSRetrieveAuthGroup( IN ULONGLONG ProcessId,
                           ThreadId);
 
             AFSReleaseResource( pDeviceExt->Specific.Control.ProcessTree.TreeLock);
+
             try_return( ntStatus);
         }
 
@@ -224,7 +225,8 @@ AFSRetrieveAuthGroup( IN ULONGLONG ProcessId,
                           ProcessId,
                           ThreadId);
 
-            pAuthGroup = AFSValidateProcessEntry();
+            pAuthGroup = AFSValidateProcessEntry( PsGetCurrentProcessId(),
+                                                  FALSE);
 
             if( pAuthGroup != NULL)
             {
@@ -427,7 +429,6 @@ AFSCreateSetProcessAuthGroup( AFSAuthGroupRequestCB *CreateSetAuthGroup)
     AFSProcessAuthGroupCB *pAuthGroup = NULL, *pLastAuthGroup = NULL;
     ULONG ulSessionId = 0;
     ULONGLONG ullTableHash = 0;
-    GUID stAuthGroup;
     UNICODE_STRING uniCallerSID;
     BOOLEAN bImpersonation = FALSE;
 
@@ -1062,7 +1063,6 @@ AFSResetActiveProcessAuthGroup( IN IN AFSAuthGroupRequestCB *AuthGroup)
 {
 
     NTSTATUS ntStatus = STATUS_SUCCESS;
-    GUID *pAuthGroup = NULL;
     AFSProcessCB *pProcessCB = NULL;
     AFSThreadCB *pThreadCB = NULL;
     AFSDeviceExt *pDeviceExt = (AFSDeviceExt *)AFSDeviceObject->DeviceExtension;
@@ -1182,7 +1182,6 @@ AFSCreateAuthGroupForSIDorLogonSession( IN AFSAuthGroupRequestCB *AuthGroupReque
     AFSSIDEntryCB *pSIDEntryCB = NULL;
     ULONG ulSessionId = 0;
     ULONGLONG ullTableHash = 0;
-    GUID stAuthGroup;
     UNICODE_STRING uniCallerSID;
     UNICODE_STRING uniGUID;
     BOOLEAN bLocalSystem = FALSE;
