@@ -55,7 +55,7 @@ AFSSetSecurity( IN PDEVICE_OBJECT DeviceObject,
 
         AFSDbgLogMsg( AFS_SUBSYSTEM_FILE_PROCESSING,
                       AFS_TRACE_LEVEL_VERBOSE,
-                      "AFSSetSecurity Entry for FO %08lX\n",
+                      "AFSSetSecurity Entry for FO %p\n",
                       pIrpSp->FileObject);
 
         if( DeviceObject == AFSDeviceObject)
@@ -120,12 +120,14 @@ try_exit:
 
         NOTHING;
     }
-    __except( AFSExceptionFilter( GetExceptionCode(), GetExceptionInformation()) )
+    __except( AFSExceptionFilter( __FUNCTION__, GetExceptionCode(), GetExceptionInformation()) )
     {
 
         AFSDbgLogMsg( 0,
                       0,
                       "EXCEPTION - AFSSetSecurity\n");
+
+        AFSDumpTraceFilesFnc();
     }
 
     return ntStatus;
@@ -148,7 +150,7 @@ AFSQuerySecurity( IN PDEVICE_OBJECT DeviceObject,
 
         AFSDbgLogMsg( AFS_SUBSYSTEM_FILE_PROCESSING,
                       AFS_TRACE_LEVEL_VERBOSE,
-                      "AFSQuerySecurity Entry for FO %08lX\n",
+                      "AFSQuerySecurity Entry for FO %p\n",
                       pIrpSp->FileObject);
 
         if( DeviceObject == AFSDeviceObject)
@@ -213,12 +215,14 @@ try_exit:
 
         NOTHING;
     }
-    __except( AFSExceptionFilter( GetExceptionCode(), GetExceptionInformation()) )
+    __except( AFSExceptionFilter( __FUNCTION__, GetExceptionCode(), GetExceptionInformation()) )
     {
 
         AFSDbgLogMsg( 0,
                       0,
                       "EXCEPTION - AFSQuerySecurity\n");
+
+        AFSDumpTraceFilesFnc();
     }
 
     return ntStatus;
