@@ -45,7 +45,6 @@ AFSLocateCaseSensitiveDirEntry( IN AFSDirectoryCB *RootNode,
 {
 
     NTSTATUS    ntStatus = STATUS_SUCCESS;
-    AFSDirectoryCB   *pEntry = NULL;
     AFSDirectoryCB   *pCurrentEntry = NULL;
 
     pCurrentEntry = RootNode;
@@ -164,7 +163,6 @@ AFSLocateCaseInsensitiveDirEntry( IN AFSDirectoryCB *RootNode,
 {
 
     NTSTATUS    ntStatus = STATUS_SUCCESS;
-    AFSDirectoryCB   *pEntry = NULL;
     AFSDirectoryCB   *pCurrentEntry = NULL;
 
     pCurrentEntry = RootNode;
@@ -880,7 +878,6 @@ AFSLocateShortNameDirEntry( IN AFSDirectoryCB *RootNode,
 {
 
     NTSTATUS    ntStatus = STATUS_SUCCESS;
-    AFSDirectoryCB   *pEntry = NULL;
     AFSDirectoryCB   *pCurrentEntry = NULL;
 
     pCurrentEntry = RootNode;
@@ -1255,8 +1252,7 @@ AFSLocateHashEntry( IN AFSBTreeEntry *TopNode,
                     IN OUT AFSBTreeEntry **TreeEntry)
 {
 
-    NTSTATUS         ntStatus = STATUS_SUCCESS;
-    AFSBTreeEntry   *pEntry = NULL;
+    NTSTATUS         ntStatus = STATUS_NOT_FOUND;
     AFSBTreeEntry   *pCurrentEntry = NULL;
 
     pCurrentEntry = TopNode;
@@ -1283,7 +1279,7 @@ AFSLocateHashEntry( IN AFSBTreeEntry *TopNode,
 
             *TreeEntry = TopNode;
 
-            try_return( ntStatus);
+            try_return( ntStatus = STATUS_SUCCESS);
         }
 
         //
@@ -1354,6 +1350,8 @@ AFSLocateHashEntry( IN AFSBTreeEntry *TopNode,
 
                 *TreeEntry = pCurrentEntry;
 
+                ntStatus = STATUS_SUCCESS;
+
                 break;
             }
         }
